/* Different SS7 version */
#define SIP_ISUP_ITU	1
#define SIP_ISUP_ANSI	2
#define SIP_ISUP_ETSI	3
#define SIP_ISUP_ISO	4

/* ISUP messages */
#define SIP_IAM	0x01
#define SIP_SAM	0x02
#define SIP_INR	0x03
#define SIP_INF	0x04
#define SIP_COT	0x05
#define SIP_ACM	0x06
#define SIP_CON	0x07
#define SIP_FOT	0x08
#define SIP_ANM	0x09
#define SIP_REL	0x0c
#define SIP_SUS	0x0d
#define SIP_RES	0x0e
#define SIP_RLC	0x10
#define SIP_CCR	0x11
#define SIP_RSC	0x12
#define SIP_BLO	0x13
#define SIP_UBL	0x14
#define SIP_BLA	0x15
#define SIP_UBA	0x16
#define SIP_GRS	0x17
#define SIP_CGB	0x18
#define SIP_CGU	0x19
#define SIP_CGBA	0x1a
#define SIP_CGUA	0x1b
#define SIP_CMR	0x1c
#define SIP_CMC	0x1d
#define SIP_CMRJ	0x1e
#define SIP_FAR	0x1f
#define SIP_FAA	0x20
#define SIP_FRJ	0x21
#define SIP_FAD	0x22
#define SIP_FAI	0x23
#define SIP_LPA	0x24
#define SIP_CSVR	0x25
#define SIP_CSVS	0x26
#define SIP_DRS	0x27
#define SIP_PAM	0x28
#define SIP_GRA	0x29
#define SIP_CQM	0x2a
#define SIP_CQR	0x2b
#define SIP_CPG	0x2c
#define SIP_USR	0x2d
#define SIP_UCIC	0x2e
#define SIP_CFN	0x2f
#define SIP_OLM	0x30
#define SIP_CRG	0x31
#define SIP_FAC	0x33
#define SIP_CRA	0xe9
#define SIP_CRM	0xea
#define SIP_CVR	0xeb
#define SIP_CVT	0xec
#define SIP_EXM	0xed

/* ISUP Parameters */
#define SIP_PARM_NATURE_OF_CONNECTION_IND 0x06
#define SIP_PARM_FORWARD_CALL_IND 0x07
#define SIP_PARM_CALLING_PARTY_CAT 0x09
#define SIP_PARM_USER_SERVICE_INFO 0x1d
#define SIP_PARM_TRANSMISSION_MEDIUM_REQS 0x02
#define SIP_PARM_CALLED_PARTY_NUM 0x04
#define SIP_PARM_ACCESS_TRANS 0x03
#define SIP_PARM_BUSINESS_GRP 0xc6
#define SIP_PARM_CALL_REF 0x01
#define SIP_PARM_CALLING_PARTY_NUM 0x0a
#define SIP_PARM_CARRIER_ID 0xc5
#define SIP_PARM_SELECTION_INFO 0xee
#define SIP_PARM_CHARGE_NUMBER 0xeb
#define SIP_PARM_CIRCUIT_ASSIGNMENT_MAP 0x25
#define SIP_PARM_OPT_BACKWARD_CALL_IND 0x29
#define SIP_PARM_CONNECTION_REQ 0x0d
#define SIP_PARM_CONTINUITY_IND 0x10
#define SIP_PARM_CUG_INTERLOCK_CODE 0x1c
#define SIP_PARM_EGRESS_SERV 0xc3
#define SIP_PARM_GENERIC_ADDR 0xc0
#define SIP_PARM_GENERIC_DIGITS 0xc1
#define SIP_PARM_GENERIC_NAME 0xc7
#define SIP_PARM_GENERIC_NOTIFICATION_IND 0x2c
#define SIP_PARM_BACKWARD_CALL_IND 0x11
#define SIP_PARM_CAUSE 0x12
#define SIP_PARM_CIRCUIT_GROUP_SUPERVISION_IND 0x15
#define SIP_PARM_RANGE_AND_STATUS 0x16
#define SIP_PARM_PROPAGATION_DELAY 0x31
#define SIP_PARM_EVENT_INFO 0x24
#define SIP_PARM_HOP_COUNTER 0x3d
#define SIP_PARM_OPT_FORWARD_CALL_INDICATOR 0x08
#define SIP_PARM_LOCATION_NUMBER 0x3f
#define SIP_PARM_ORIG_LINE_INFO 0xea
#define SIP_PARM_REDIRECTION_INFO 0x13
#define SIP_PARM_ORIGINAL_CALLED_NUM 0x28
#define SIP_PARM_JIP 0xc4
#define SIP_PARM_ECHO_CONTROL_INFO 0x37
#define SIP_PARM_PARAMETER_COMPAT_INFO 0x39
#define SIP_PARM_CIRCUIT_STATE_IND 0x26
#define SIP_PARM_TRANSIT_NETWORK_SELECTION 0x23
#define SIP_PARM_LOCAL_SERVICE_PROVIDER_IDENTIFICATION 0xe4
#define SIP_PARM_FACILITY_IND 0x18
#define SIP_PARM_REDIRECTING_NUMBER 0x0b 
#define SIP_PARM_ACCESS_DELIVERY_INFO 0x2e

#define PARM_TYPE_FIXED		0x01
#define PARM_TYPE_VARIABLE	0x02
#define PARM_TYPE_OPTIONAL	0x03

#define ISUP_MAX_NUM 64
/* From GR-317 for the generic name filed: 15 + 1 */
#define ISUP_MAX_NAME 16

/* ISUP Buffering */
struct sip_isup_buffer
{       int version;    /* ITU 1, ANSI 2,ETSI 3, ISO  4 */
        int message_length;
        unsigned char *message;
};

static char digit2char(unsigned char digit)
{
	switch (digit & 0xf) {
		case 0:
			return '0';
		case 1:
			return '1';
		case 2:
			return '2';
		case 3:
			return '3';
		case 4:
			return '4';
		case 5:
			return '5';
		case 6:
			return '6';
		case 7:
			return '7';
		case 8:
			return '8';
		case 9:
			return '9';
		case 15:
			return '#';
		default:
			return 0;
	}
}

static void isup_get_number(char *dest, unsigned char *src, int srclen, int oddeven)
{
	int i;

	if (oddeven < 2) {
		/* BCD odd or even */
		for (i = 0; i < ((srclen * 2) - oddeven); i++)
			dest[i] = digit2char(src[i/2] >> ((i % 2) ? 4 : 0));
	} else {
		/* oddeven = 2 for IA5 characters */
		for (i = 0; i < srclen; i++)
			dest[i] = src[i];
	}
	dest[i] = '\0'; 
}

static char *int2str(int to_convert)
{ char	buffer[32];
  char *buffer_string;
	int len=0;
	sprintf(buffer,"%d",to_convert);
	len=strlen(buffer)+1;
	buffer_string=malloc(len);
	memcpy(buffer_string,buffer,len);
	buffer_string[len]='\0';
	return buffer_string;
}

#define FUNC_DECODE(name) int ((name))(struct sip_pvt *p, struct sip_isup_buffer buffer, int messagetype, unsigned char *parm, int len)
static FUNC_DECODE(nature_of_connection_ind)
{
	unsigned char con = parm[0];
	char *continuity;
	
	pbx_builtin_setvar_helper(p->owner, "satellites_in_connection", int2str(con&0x03));
	con>>=2; 
	switch (con & 0x03) {
		case 0:
			continuity = "Check not required";
			break;
		case 1:
			continuity = "Check required on this circuit";
			break;
		case 2:
			continuity = "Check performed on a previous circuit";
			break;
		case 3:
			continuity = "spare";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "continuity_check", continuity);

	con>>=2;
	con &= 0x01;

	pbx_builtin_setvar_helper(p->owner, "echo_control_device", con ? "included" : "not included");

	return 1;
}

static FUNC_DECODE(forward_call_ind)
{
	char *cb_str, *hg_str=NULL, *kj_str;
	switch ((parm[0] >> 1) & 3) {
		case 0:
			cb_str = "no end-to-end";
			break;
		case 1:
			cb_str = "pass-along";
			break;
		case 2:
			cb_str = "SCCP";
			break;
		case 3:
			cb_str = "pass-along and SCCP";
			break;
	}

	switch ((parm[0] >> 6) & 3) {
		case 0:
			hg_str = "ISDN user part preferred all the way";
			break;
		case 1:
			hg_str = "ISDN user part not preferred all the way";
			break;
		case 2:
			hg_str = "ISDN user part requried all the way";
			break;
		case 3:
			hg_str = "spare";
			break;
	}

	switch ((parm[1] >> 1) & 3) {
		case 0:
			kj_str = "no indication";
			break;
		case 1:
			kj_str = "connectionless method available";
			break;
		case 2:
			kj_str = "connection oriented method available";
			break;
		case 3:
			kj_str = "connectionless and connection oriented method available";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "national_call",  (parm[0] & 1) ? "international" : "national");
	pbx_builtin_setvar_helper(p->owner, "end_to_end_method", cb_str);	
	pbx_builtin_setvar_helper(p->owner, "interworking", ((parm[0] >> 3) & 1) ? "" : "no ");
	pbx_builtin_setvar_helper(p->owner, "end_to_end_information", ((parm[0]>>4)&1) ? "" : "no ");
	pbx_builtin_setvar_helper(p->owner, "isup_all_the_way",((parm[0]>>5)&1) ? "" : "not ");
	pbx_builtin_setvar_helper(p->owner, "isup_preference",hg_str);
	pbx_builtin_setvar_helper(p->owner, "isdn_access_origination",(parm[1] & 1) ? "ISDN" : "non-ISDN");
	pbx_builtin_setvar_helper(p->owner, "sccp_method",kj_str);
	return 2;
}

static FUNC_DECODE(calling_party_cat)
{
	char *cattype;

	switch (parm[0]) {
		case 1:
			cattype = "Operator, French";
			break;
		case 2:
			cattype = "Operator, English";
			break;
		case 3:
			cattype = "Operator, German";
			break;
		case 4:
			cattype = "Operator, Russian";
			break;
		case 5:
			cattype = "Operator, Spanish";
			break;
		case 9:
			cattype = "Reserved";
			break;
		case 10:
			cattype = "Ordinary calling subscriber";
			break;
		case 11:
			cattype = "Calling subscriber with priority";
			break;
		case 12:
			cattype = "Data Call (voice band data)";
			break;
		case 13:
			cattype = "Test Call";
			break;
		case 15:
			cattype = "Payphone";
			break;
		default:
			cattype = "Unknown";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "calling_party_category",cattype);
	return 1;
}

static FUNC_DECODE(transmission_medium_reqs)
{
	char *type;

	switch (parm[0]) {
		case 0:
			type = "Speech";
			break;
		case 1:
			type = "Spare";
			break;
		case 2:
			type = "64 kbit/s unrestricted";
			break;
		case 4:
			type = "3.1 khz audio";
			break;
		case 6:
			type = "64 kbit/s preferred";
			break;
		case 7:
			type = "2 x 64 kbit/s unrestricted";
			break;
		case 8:
			type = "384 kbit/s unrestricted";
			break;
		case 9:
			type = "1536 kbit/s unrestricted";
			break;
		case 10:
			type = "1920 kbit/s unrestricted";
			break;
		default:
			type = "N x 64kbit/s unrestricted or possibly spare";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "transmission_medium_req",type);
	return 1;
}

static FUNC_DECODE(called_party_num)
{
	int oddeven = (parm[0] >> 7) & 0x1;
	char numbuf[64] = "";
 	pbx_builtin_setvar_helper(p->owner, "called_nature_of_address", int2str(parm[0] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "called_ni", int2str((parm[1] >> 7) & 0x1));
	pbx_builtin_setvar_helper(p->owner, "called_numbering_plan", int2str((parm[1] >> 4) & 0x7));
	isup_get_number(numbuf, &parm[2], len - 2, oddeven);
	pbx_builtin_setvar_helper(p->owner, "called_party_number",numbuf);

	return len;
}

static FUNC_DECODE(cause)
{
	char *cause;
	switch (parm[1] & 0x7f) {
		case 1:
			cause = "Unallocated (unassigned) number";
			break;
		case 2:
			cause = "No route to specified transit network";
			break;
		case 3:
			cause = "No route to destination";
			break;
		case 4:
			cause = "Send special information tone";
			break;
		case 5:
			cause = "Misdialled trunk prefix";
			break;
		case 6:
			cause = "Channel unacceptable";
			break;
		case 7:
			cause = "Call awarded and being delivered in an established channel";
			break;
		case 8:
			cause = "Preemption";
			break;
		case 9:
			cause = "Preemption - circuit reserved for reuse";
			break;
		case 16:
			cause = "Normal call clearing";
			break;
		case 17:
			cause = "User busy";
			break;
		case 18:
			cause = "No user responding";
			break;
		case 19:
			cause = "No answer from user (user alerted)";
			break;
		case 20:
			cause = "Subscriber absent";
			break;
		case 21:
			cause = "Call rejected";
			break;
		case 22:
			cause = "Number changed";
			break;
		case 23:
			cause = "Redirection to new destination";
			break;
		case 25:
			cause = "Exchange routing error";
			break;
		case 26:
			cause = "Non-selected user clearing";
			break;
		case 27:
			cause = "Destination out of order";
			break;
		case 28:
			cause = "Invalid number format (address incomplete)";
			break;
		case 29:
			cause = "Facility rejected";
			break;
		case 30:
			cause = "Response to STATUS ENQUIRY";
			break;
		case 31:
			cause = "Normal, unspecified";
			break;
		case 34:
			cause = "No circuit/channel available";
			break;
		case 38:
			cause = "Network out of order";
			break;
		case 39:
			cause = "Permanent frame mode connection out of service";
			break;
		case 40:
			cause = "Permanent frame mode connection operational";
			break;
		case 41:
			cause = "Temporary failure";
			break;
		case 42:
			cause = "Switching equipment congestion";
			break;
/* TODO: Finish the rest of these */
		default:
			cause = "Unknown";
	}
	pbx_builtin_setvar_helper(p->owner, "coding_standard", int2str((parm[0] >> 5) & 3));
	pbx_builtin_setvar_helper(p->owner, "location", int2str((parm[0] >> 5) & 3));
	pbx_builtin_setvar_helper(p->owner, "cause_class", int2str((parm[1]>>4) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "cause_sub_class", int2str(parm[1] & 0xf));
	pbx_builtin_setvar_helper(p->owner, "cause", cause);

	return len;
}

static FUNC_DECODE(continuity_ind)
{	pbx_builtin_setvar_helper(p->owner, "opt_continuity_check",(0x01 & parm[0]) ? "successful" : "failed");

	return 1;
}

static FUNC_DECODE(access_transport)
{
	return len;	
}

static FUNC_DECODE(call_ref)
{
	unsigned int ptc, callr;
	
	callr = parm[0] | (parm[1] << 8) | (parm[2] << 16);
	if (buffer.version == SIP_ISUP_ANSI)
		ptc = parm[3] | (parm[4] << 8) | (parm[5] << 16);
	else
		ptc = parm[3] | (parm[4] << 8);
	
	pbx_builtin_setvar_helper(p->owner, "call_reference", int2str(callr));
	pbx_builtin_setvar_helper(p->owner, "point_code", int2str(ptc));
	
	return len;
}

static FUNC_DECODE(calling_party_num)
{
	int oddeven = (parm[0] >> 7) & 0x1;
	char numbuf[64] = "";
	pbx_builtin_setvar_helper(p->owner, "calling_nature_of_address", int2str(parm[0] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "calling_ni", int2str((parm[1] >> 7) & 0x1));
	pbx_builtin_setvar_helper(p->owner, "calling_numbering_plan", int2str((parm[1] >> 4) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "calling_presentation", int2str((parm[1] >> 2) & 0x3));
	pbx_builtin_setvar_helper(p->owner, "calling_screening", int2str(parm[1] & 0x3));

	isup_get_number(numbuf, &parm[2], len - 2, oddeven);
	pbx_builtin_setvar_helper(p->owner, "calling_party_number", numbuf);

	return len;
}

static FUNC_DECODE(carrier_identification)
{
	return len;
}

static FUNC_DECODE(charge_number)
{
	int oddeven = (parm[0] >> 7) & 0x1;
	char numbuf[64] = "";
	pbx_builtin_setvar_helper(p->owner, "cn_nature_of_address", int2str(parm[0] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "cn_numbering_plan", int2str((parm[1] >> 4) & 0x7));
	isup_get_number(numbuf, &parm[2], len - 2, oddeven);
	pbx_builtin_setvar_helper(p->owner, "charge_number", numbuf);

	return len;
}

struct parm_func {
	int parm;
	char *name;
	FUNC_DECODE(*decode);
};

static FUNC_DECODE(generic_address)
{
	int oddeven = (parm[1] >> 7) & 0x1;
	char numbuf[64] = "";
	pbx_builtin_setvar_helper(p->owner, "ga_type_of_address", int2str(parm[0]));
	pbx_builtin_setvar_helper(p->owner, "ga_nature_of_address", int2str(parm[1] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "ga_odd_even", int2str((parm[1] >> 7) & 0x1));
	pbx_builtin_setvar_helper(p->owner, "ga_reserved", int2str(parm[2] & 0x3));
	pbx_builtin_setvar_helper(p->owner, "ga_presentation", int2str((parm[2] >> 2) & 0x3));
	pbx_builtin_setvar_helper(p->owner, "ga_numbering_plan", int2str((parm[2] >> 4) & 0x7));

	isup_get_number(numbuf, &parm[3], len - 3, oddeven);
	pbx_builtin_setvar_helper(p->owner, "generic_address",numbuf);	
	
	return len;
}

static FUNC_DECODE(generic_digits)
{
	int oddeven = (parm[0] >> 5) & 0x7;
	char numbuf[64] = "";

	pbx_builtin_setvar_helper(p->owner, "gd_type_of_digits", int2str(parm[0] & 0x1f));
	pbx_builtin_setvar_helper(p->owner, "gd_encoding_scheme", int2str((parm[0] >> 5) & 0x7));
	isup_get_number(numbuf, &parm[1], len - 1, oddeven);
	pbx_builtin_setvar_helper(p->owner, "generic_digits", numbuf);

	return len;
	
}

static FUNC_DECODE(generic_name)
{
	unsigned int typeofname = (parm[0] >> 5) & 0x7;
	unsigned int avail = (parm[0] >> 4) & 0x1;
	unsigned int presentation = parm[0] & 0x3;
	char name[ISUP_MAX_NAME + 1];

	memcpy(name, &parm[1], len - 1);
	pbx_builtin_setvar_helper(p->owner, "gn_type_of_name",(typeofname == 1) ? "Calling Name" : "Unknown");
	pbx_builtin_setvar_helper(p->owner, "gn_available", (avail == 1) ? "Name not available" : "Name available, or availability unknown");
	pbx_builtin_setvar_helper(p->owner, "gn_presentation", int2str(presentation));
	pbx_builtin_setvar_helper(p->owner, "generic_name", name);

	return len;
}

static FUNC_DECODE(tns)
{	pbx_builtin_setvar_helper(p->owner, "tns_type_of_network", int2str((parm[0] >> 4) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "tns_network_id_plan", int2str(parm[0] & 0xf));
	pbx_builtin_setvar_helper(p->owner, "tns_network_id1", int2str(parm[1]));
	pbx_builtin_setvar_helper(p->owner, "tns_network_id2", int2str(parm[2]));
	pbx_builtin_setvar_helper(p->owner, "tns_circuit_code", int2str((parm[3] >> 4) & 0xf));
	
	return len;
}

static FUNC_DECODE(propagation_delay_cntr)
{	pbx_builtin_setvar_helper(p->owner, "propagation_delay_cntr", int2str((unsigned short)(((parm[0] & 0xff) << 8) | (parm[1] & 0xff))));
	return len;
}

static FUNC_DECODE(hop_counter)
{
	return 1;
}

static FUNC_DECODE(backward_call_ind)
{
	unsigned char ba = parm[0] & 0x3;
	unsigned char dc = (parm[0] >> 2) & 0x3;
	unsigned char fe = (parm[0] >> 4) & 0x3;
	unsigned char hg = (parm[0] >> 6) & 0x3;
	unsigned char i = parm[1] & 0x1;
	unsigned char j = (parm[1] >> 1) & 0x1;
	unsigned char k = (parm[1] >> 2) & 0x1;
	unsigned char l = (parm[1] >> 3) & 0x1;
	unsigned char m = (parm[1] >> 4) & 0x1;
	unsigned char n = (parm[1] >> 5) & 0x1;
	unsigned char pq = (parm[1] >> 7) & 0x3;

	pbx_builtin_setvar_helper(p->owner, "bc_charge", int2str(ba)); /* Backward Call ind */
	pbx_builtin_setvar_helper(p->owner, "bc_called_party_status", int2str(dc));
	pbx_builtin_setvar_helper(p->owner, "bc_called_party_category", int2str(fe));
	pbx_builtin_setvar_helper(p->owner, "bc_end_to_end_method", int2str(hg));
	pbx_builtin_setvar_helper(p->owner, "bc_interworking", int2str(i));
	pbx_builtin_setvar_helper(p->owner, "bc_end_to_end_info", int2str(j));
	pbx_builtin_setvar_helper(p->owner, "bc_isup", int2str(k));
	pbx_builtin_setvar_helper(p->owner, "bc_holding", int2str(l));
	pbx_builtin_setvar_helper(p->owner, "bc_isdc_acces", int2str(m));
	pbx_builtin_setvar_helper(p->owner, "bc_echo_control_device", int2str(n));
	pbx_builtin_setvar_helper(p->owner, "bc_sccp_method", int2str(pq));

	return 2;
}

static FUNC_DECODE(opt_backward_call_ind)
{
	unsigned char a, b, c, d;
	a = parm[0] & 1;
	b = (parm[0] >> 1) & 1;
	c = (parm[0] >> 2) & 1;
	d = (parm[0] >> 3) & 1;

	pbx_builtin_setvar_helper(p->owner, "obc_inband", int2str(a)); /* Optional Backward Call ind */
	pbx_builtin_setvar_helper(p->owner, "obc_diversion_may_occur", int2str(b));
	pbx_builtin_setvar_helper(p->owner, "obc_simple_segmentation", int2str(c));
	pbx_builtin_setvar_helper(p->owner, "obc_mlpp_user", int2str(c));

	return 1;
}

static FUNC_DECODE(circuit_group_supervision)
{
	char *name;

	switch (parm[0] & 0x3) {
	case 0:
		name = "Maintenance oriented";
		break;
	case 1:
		name = "Hardware Failure oriented";
		break;
	case 2:
		name = "Reserved for national use";
		break;
	case 3:
		name = "Spare";
		break;
	default:
		name = "Huh?!";
	}
	pbx_builtin_setvar_helper(p->owner, "circuit_group_supervision", name);

	return 1;
}

static FUNC_DECODE(range_and_status)
{	pbx_builtin_setvar_helper(p->owner, "isup_range_status", int2str(parm[0] & 0xff));
	return len;
}

static FUNC_DECODE(event_info)
{
	char *name;

	switch (parm[0]) {
		case 0:
			name = "spare";
			break;
		case 1:
			name = "ALERTING";
			break;
		case 2:
			name = "PROGRESS";
			break;
		case 3:
			name = "In-band information or an appropriate pattern is now available";
			break;
		case 4:
			name = "Call forward on busy";
			break;
		case 5:
			name = "Call forward on no reply";
			break;
		case 6:
			name = "Call forward unconditional";
			break;
		default:
			name = "Spare";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "event_info", name);
	return 1;
}

static FUNC_DECODE(originating_line_information)
{
	char *name;

	switch (parm[0]) {
		case 0:
			name = " Plain Old Telephone Service POTS";
			break;
		case 1:
			name = " Multiparty line";
			break;
		case 2:
			name = " ANI Failure";
			break;
		case 3:
		case 4:
		case 5:
			name = " Unassigned";
			break;
		case 6:
			name = " Station Level Rating";
			break;
		case 7:
			name = " Special Operator Handling Required";
			break;
		case 8:
		case 9:
			name = "Unassigned";
			break;
		case 10:
			name = "Not assignable";
			break;
		case 11:
			name = "Unassigned";
			break;
		case 12:
		case 13:
		case 14:
		case 15:
		case 16:
		case 17:
		case 18:
		case 19:
			name = "Not assignable";
			break;
		case 20:
			name = "Automatic Identified Outward Dialing";
			break;
		case 21:
		case 22:
			name = "Unassigned";
			break;
		case 23:
			name = "Coin or Non-Coin";
			break;
		case 24:
		case 25:
			name = "Toll Free Service translated to POTS";
			break;
		case 26:
			name = "Unassigned";
			break;
		case 27:
			name = "Pay Station using Coin Control Signalling";
			break;
		case 28:
			name = "Unassigned";
			break;
		case 29:
			name = "Prison/Inmate Service";
			break;
		case 30:
		case 31:
		case 32:
			name = "Intercept";
			break;
		case 33:
			name = "Unassigned";
			break;
		case 34:
			name = "Telco Operator Handled Call";
			break;
		case 35:
		case 36:
		case 37:
		case 38:
		case 39:
			name = "Unassigned";
			break;
		case 40:
		case 41:
		case 42:
		case 43:
		case 44:
		case 45:
		case 46:
		case 47:
		case 48:
		case 49:
			name = "Unrestricted Use - locally determined by carrier";
			break;
		case 50:
		case 51:
			name = "Unassigned";
			break;
		case 52:
			name = "OUTWATS";
			break;
		case 53:
		case 54:
		case 55:
		case 56:
		case 57:
		case 58:
		case 59:
			name = "Unassigned";
			break;
		case 60:
			name = "TRS Unrestricted Line";
			break;
		case 61:
			name = "Cellular Wireless PCS Type 1";
			break;
		case 62:
			name = "Cellular Wireless PCS Type 2";
			break;
		case 63:
			name = "Cellular Wireless PCS Roaming";
			break;
		case 64:
		case 65:
			name = "Unassigned";
			break;
		case 66:
			name = "TRS Hotel Motel";
			break;
		case 67:
			name = "TRS Restricted Line";
			break;
		case 68:
		case 69:
			name = "Unassigned";
			break;
		case 70:
			name = "Pay Station No network Coin Control Signalling";
			break;
		case 71:
		case 72:
		case 73:
		case 74:
		case 75:
		case 76:
		case 77:
		case 78:
		case 79:
			name = "Unassigned";
			break;
		case 80:
		case 81:
		case 82:
		case 83:
		case 84:
		case 85:
		case 86:
		case 87:
		case 88:
		case 89:
			name = "Reserved";
			break;
		case 90:
		case 91:
		case 92:
			name = "Unassigned";
			break;
		case 93:
			name = "Private Virtual Network Type of service call";
			break;
		case 94:
		case 95:
		case 96:
		case 97:
		case 98:
		case 99:
			name = "Unassigned";
			break;

		default:
			name = "Unknown to Asterisk ";
	}
	pbx_builtin_setvar_helper(p->owner, "oli", name);

	return 1;
}

static FUNC_DECODE(redirection_info)
{
	char *redirect_ind, *orig_redir_reas, *redir_reas=NULL;

	switch (parm[0] & 0x7) {
		case 0:
			redirect_ind = "No Redirection (national use)";
			break;
		case 1:
			redirect_ind = "Call rerouted (national use)";
			break;
		case 2:
			redirect_ind = "Call rerouted, all rediection information presentation restricted (national use)";
			break;
		case 3:
			redirect_ind = "Call diverted";
			break;
		case 4:
			redirect_ind = "Call diverted, all redirection information presentation restricted";
			break;
		case 5:
			redirect_ind = "Call rerouted, redirection number presentation restricted (national use)";
			break;
		case 6:
			redirect_ind = "Call diversion, redirection number presentation restricted (national use)";
			break;
		case 7:
			redirect_ind = "spare";
			break;
		default:
			redirect_ind = "Unknown";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "redirecting_indicator", redirect_ind);

	switch ((parm[0] >> 4) & 0xf) {
		case 0:
			orig_redir_reas = "Unknown/not available";
			break;
		case 1:
			orig_redir_reas = "User busy (national use)";
			break;
		case 2:
			orig_redir_reas = "No reply (national use)";
			break;
		case 3:
			orig_redir_reas = "Unconditional (national use)";
			break;
		default:
			orig_redir_reas = "spare";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "original_redirecting_reason", orig_redir_reas);
	pbx_builtin_setvar_helper(p->owner, "redirecting_counter", int2str(parm[1] & 0x7));

	switch ((parm[1] >> 4) & 0xf) {
		case 0:
			orig_redir_reas = "Unknown/not available";
			break;
		case 1:
			redir_reas = "User busy";
			break;
		case 2:
			redir_reas = "No reply";
			break;
		case 3:
			redir_reas = "Unconditional";
			break;
		case 4:
			redir_reas = "Deflection during alerting";
			break;
		case 5:
			redir_reas = "Deflection immediate response";
			break;
		case 6:
			redir_reas = "Mobile subscriber not reachable";
			break;
		default:
			redir_reas = "spare";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "redirecting_reason", redir_reas);

	return 2;
}

static FUNC_DECODE(original_called_num)
{
	int oddeven = (parm[0] >> 7) & 0x1;
	char numbuf[64] = "";

	pbx_builtin_setvar_helper(p->owner, "ocn_nature_of_addres", int2str(parm[0] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "ocn_numbering_plan", int2str((parm[1] >> 4) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "ocn_presentation", int2str((parm[1] >> 2) & 0x3));

	isup_get_number(numbuf, &parm[2], len - 2, oddeven);
	pbx_builtin_setvar_helper(p->owner, "original_called_num", numbuf);
	
	return len;
}

static FUNC_DECODE(jip)
{
	char numbuf[64] = "";
	
	isup_get_number(numbuf, &parm[0], len , 0);
	pbx_builtin_setvar_helper(p->owner, "jurisdiction_info", numbuf);
	return len;
}

static FUNC_DECODE(echo_control_info)
{
	unsigned char ba = parm[0] & 0x3;
	unsigned char dc = (parm[0] >> 2) & 0x3;
	unsigned char fe = (parm[0] >> 4) & 0x3;
	unsigned char hg = (parm[0] >> 6) & 0x3;
	char *ba_str, *dc_str, *fe_str, *hg_str;

	switch (ba) {
		case 0:
			ba_str = "no information";
			break;
		case 1:
			ba_str = "outgoing echo control device not included and not available";
			break;
		case 2:
			ba_str = "outgoing echo control device included";
			break;
		case 3:
			ba_str = "outgoing echo control device not included but available";
			break;
		default:
			ba_str = "unknown";
			break;
	}

	switch (dc) {
		case 0:
			dc_str = "no information";
			break;
		case 1:
			dc_str = "incoming echo control device not included and not available";
			break;
		case 2:
			dc_str = "incoming echo control device included";
			break;
		case 3:
			dc_str = "incoming echo control device not included but available";
			break;
		default:
			dc_str = "unknown";
			break;
	}

	switch (fe) {
		case 0:
			fe_str = "no information";
			break;
		case 1:
			fe_str = "outgoing echo control device activation request";
			break;
		case 2:
			fe_str = "outgoing echo control device deactivation request";
			break;
		case 3:
			fe_str = "spare";
			break;
		default:
			fe_str = "unknown";
			break;
	}

	switch (hg) {
		case 0:
			hg_str = "no information";
			break;
		case 1:
			hg_str = "incoming echo control device activation request";
			break;
		case 2:
			hg_str = "incoming echo control device deactivation request";
			break;
		case 3:
			hg_str = "spare";
			break;
		default:
			hg_str = "unknown";
			break;
	}
	pbx_builtin_setvar_helper(p->owner, "eci_outgoing_info", ba_str);
	pbx_builtin_setvar_helper(p->owner, "eci_incoming_info", dc_str);
	pbx_builtin_setvar_helper(p->owner, "eci_outgoing_request", fe_str);
	pbx_builtin_setvar_helper(p->owner, "eci_incoming_request", hg_str);

	return len;
}

static FUNC_DECODE(parameter_compat_info)
{
	return len;
}

static FUNC_DECODE(circuit_state_ind)
{
	unsigned char dcbits, babits, febits;
	char *ba_str, *dc_str, *fe_str;
	int i;
	
	for (i = 0; i < len; i++) {
		babits = parm[i] & 0x3;
		dcbits = (parm[i] >> 2) & 0x3;
		febits = (parm[i] >> 4) & 0x3;

		if (dcbits == 0) {
			switch (babits) {
				case 0:
					ba_str = "transient";
					break;
				case 1:
				case 2:
					ba_str = "spare";
					break;
				case 3:
					ba_str = "unequipped";
					break;
			}
		} else {
			switch (babits) {
				case 0:
					ba_str = "no blocking (active)";
					break;
				case 1:
					ba_str = "locally blocked";
					break;
				case 2:
					ba_str = "remotely blocked";
					break;
				case 3:
					ba_str = "locally and remotely blocked";
					break;
			}

			switch (dcbits) {
				case 1:
					dc_str = "circuit incoming busy";
					break;
				case 2:
					dc_str = "circuit outgoing busy";
					break;
				case 3:
					dc_str = "idle";
					break;
			}

			switch (febits) {
				case 0:
					fe_str = "no blocking (active)";
					break;
				case 1:
					fe_str = "locally blocked";
					break;
				case 2:
					fe_str = "remotely blocked";
					break;
				case 3:
					fe_str = "locally and remotely blocked";
					break;
			}

		}
		pbx_builtin_setvar_helper(p->owner, "cs_maintenance_blocking", ba_str);
		if (!dcbits)
			continue;
		pbx_builtin_setvar_helper(p->owner, "cs_call_processing", dc_str);
                pbx_builtin_setvar_helper(p->owner, "cs_hardware_blocking", fe_str);
	}
	return len;
}

static FUNC_DECODE(lspi)
{	pbx_builtin_setvar_helper(p->owner, "lspi_type", int2str(parm[0] & 0x1f));
	pbx_builtin_setvar_helper(p->owner, "lspi_encoding_scheme", int2str((parm[0] >> 5) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "lspi_context_id", int2str(parm[1] & 0xf));
	pbx_builtin_setvar_helper(p->owner, "lspi_spare", int2str((parm[1] >> 4) & 0xf));
	pbx_builtin_setvar_helper(p->owner, "lsp_identity", int2str(parm[2]));
	
	return len;
}

static FUNC_DECODE(facility_ind)
{	pbx_builtin_setvar_helper(p->owner, "facility", int2str(parm[0]));
	return 1;
}

static FUNC_DECODE(redirecting_number)
{
	int oddeven = (parm[0] >> 7) & 0x1;
	char numbuf[64] = "";
	pbx_builtin_setvar_helper(p->owner, "rn_nature_of_address", int2str(parm[0] & 0x7f));
	pbx_builtin_setvar_helper(p->owner, "rn_ni", int2str((parm[1] >> 7) & 0x1));
	pbx_builtin_setvar_helper(p->owner, "rn_numbering_plan", int2str((parm[1] >> 4) & 0x7));
	pbx_builtin_setvar_helper(p->owner, "rn_presentation", int2str((parm[1] >> 2) & 0x3));
	pbx_builtin_setvar_helper(p->owner, "rn_screening", int2str(parm[1] & 0x3));
	
	isup_get_number(numbuf, &parm[2], len - 2, oddeven);
	pbx_builtin_setvar_helper(p->owner, "redirecting_number",numbuf);	
	
	return len;
}

static struct parm_func parms[] = {
	{SIP_PARM_NATURE_OF_CONNECTION_IND, "Nature of Connection Indicator", nature_of_connection_ind},
	{SIP_PARM_FORWARD_CALL_IND, "Forward Call Indicators", forward_call_ind},
	{SIP_PARM_CALLING_PARTY_CAT, "Calling Party's Category", calling_party_cat},
	{SIP_PARM_TRANSMISSION_MEDIUM_REQS, "Transmission Medium Requirements", transmission_medium_reqs},
	{SIP_PARM_USER_SERVICE_INFO, "User Service Information", NULL},
	{SIP_PARM_CALLED_PARTY_NUM, "Called Party Number", called_party_num},
	{SIP_PARM_CAUSE, "Cause Indicator", cause},
	{SIP_PARM_CONTINUITY_IND, "Continuity Indicator", continuity_ind},
	{SIP_PARM_ACCESS_TRANS, "Access Transport", access_transport},
	{SIP_PARM_BUSINESS_GRP, "Business Group"},
	{SIP_PARM_CALL_REF, "Call Reference", call_ref},
	{SIP_PARM_CALLING_PARTY_NUM, "Calling Party Number", calling_party_num},
	{SIP_PARM_CARRIER_ID, "Carrier Identification", carrier_identification},
	{SIP_PARM_SELECTION_INFO, "Selection Information"},
	{SIP_PARM_CHARGE_NUMBER, "Charge Number", charge_number},
	{SIP_PARM_CIRCUIT_ASSIGNMENT_MAP, "Circuit Assignment Map"},
	{SIP_PARM_CONNECTION_REQ, "Connection Request"},
	{SIP_PARM_CUG_INTERLOCK_CODE, "Interlock Code"},
	{SIP_PARM_EGRESS_SERV, "Egress Service"},
	{SIP_PARM_GENERIC_ADDR, "Generic Address", generic_address},
	{SIP_PARM_GENERIC_DIGITS, "Generic Digits", generic_digits},
	{SIP_PARM_GENERIC_NAME, "Generic Name", generic_name},
	{SIP_PARM_TRANSIT_NETWORK_SELECTION, "Transit Network Selection", tns},
	{SIP_PARM_GENERIC_NOTIFICATION_IND, "Generic Notification Indication"},
	{SIP_PARM_PROPAGATION_DELAY, "Propagation Delay Counter", propagation_delay_cntr},
	{SIP_PARM_HOP_COUNTER, "Hop Counter", hop_counter},
	{SIP_PARM_BACKWARD_CALL_IND, "Backward Call Indicator", backward_call_ind},
	{SIP_PARM_OPT_BACKWARD_CALL_IND, "Optional Backward Call Indicator", opt_backward_call_ind},
	{SIP_PARM_CIRCUIT_GROUP_SUPERVISION_IND, "Circuit Group Supervision Indicator", circuit_group_supervision},
	{SIP_PARM_RANGE_AND_STATUS, "Range and status", range_and_status},
	{SIP_PARM_EVENT_INFO, "Event Information", event_info},
	{SIP_PARM_OPT_FORWARD_CALL_INDICATOR, "Optional forward call indicator"},
	{SIP_PARM_LOCATION_NUMBER, "Location Number"},
	{SIP_PARM_ORIG_LINE_INFO, "Originating line information", originating_line_information},
	{SIP_PARM_REDIRECTION_INFO, "Redirection Information", redirection_info},
	{SIP_PARM_ORIGINAL_CALLED_NUM, "Original called number", original_called_num},
	{SIP_PARM_JIP, "Jurisdiction Information Parameter", jip},
	{SIP_PARM_ECHO_CONTROL_INFO, "Echo Control Information", echo_control_info},
	{SIP_PARM_PARAMETER_COMPAT_INFO, "Parameter Compatibility Information", parameter_compat_info},
	{SIP_PARM_CIRCUIT_STATE_IND, "Circuit State Indicator", circuit_state_ind},
	{SIP_PARM_LOCAL_SERVICE_PROVIDER_IDENTIFICATION, "Local Service Provider ID", lspi},
	{SIP_PARM_FACILITY_IND, "Facility Indicator", facility_ind},
	{SIP_PARM_REDIRECTING_NUMBER, "Redirecting Number", redirecting_number},
	{SIP_PARM_ACCESS_DELIVERY_INFO, "Access Delivery Information"}
};

static char * param2str(int parm)
{
	int x;
	int totalparms = sizeof(parms)/sizeof(struct parm_func);
	for (x = 0; x < totalparms; x++)
		if (parms[x].parm == parm)
			return parms[x].name;

	return "Unknown";
}

static void sip_isup_dump_buffer(unsigned char *data, int len)
{
	int i;

	ast_verbose("[");
	for (i = 0; i < len; i++) {
		ast_verbose("0x%x ", data[i]);
	}
	ast_verbose("]\n");
}

/* Network indicator */
#define SIP_NI_INT			0x00
#define SIP_NI_INT_SPARE	0x01
#define SIP_NI_NAT			0x02
#define SIP_NI_NAT_SPARE	0x03

/* Nature of Address Indicator */
#define SIP_NAI_SUBSCRIBER		0x01
#define SIP_NAI_UNKNOWN			0x02
#define SIP_NAI_NATIONAL		0x03
#define SIP_NAI_INTERNATIONAL	0x04

/* Charge Number Nature of Address Indicator ANSI */
#define SIP_ANI_CALLING_PARTY_SUB_NUMBER	0x01	/* ANI of the calling party; subscriber number */
#define SIP_ANI_NOTAVAIL_OR_NOTPROVIDED		0x02	/* ANI not available or not provided */
#define SIP_ANI_CALLING_PARTY_NATIONAL_NUMBER	0x03	/* ANI of the calling party; national number */
#define SIP_ANI_CALLED_PARTY_SUB_NUMBER		0x05	/* ANI of the called party; subscriber number */
#define SIP_ANI_CALLED_PARTY_NOT_PRESENT	0x06	/* ANI of the called party; no number present */
#define SIP_ANI_CALLED_PARTY_NATIONAL_NUMBER	0x07	/* ANT of the called patty; national number */

/* Address Presentation */
#define SIP_PRESENTATION_ALLOWED                       0x00
#define SIP_PRESENTATION_RESTRICTED                    0x01
#define SIP_PRESENTATION_ADDR_NOT_AVAILABLE            0x02

/* Screening */
#define SIP_SCREENING_USER_PROVIDED_NOT_VERIFIED       0x00
#define SIP_SCREENING_USER_PROVIDED                    0x01
#define SIP_SCREENING_NETWORK_PROVIDED_FAILED          0x02
#define SIP_SCREENING_NETWORK_PROVIDED                 0x03

static int sip_iam_params[] = {SIP_PARM_NATURE_OF_CONNECTION_IND, SIP_PARM_FORWARD_CALL_IND, 
	SIP_PARM_CALLING_PARTY_CAT, SIP_PARM_TRANSMISSION_MEDIUM_REQS, SIP_PARM_CALLED_PARTY_NUM,
	SIP_PARM_CALLING_PARTY_NUM, -1};

static int sip_ansi_iam_params[] = {SIP_PARM_NATURE_OF_CONNECTION_IND, SIP_PARM_FORWARD_CALL_IND,
	SIP_PARM_CALLING_PARTY_CAT, SIP_PARM_USER_SERVICE_INFO, SIP_PARM_CALLED_PARTY_NUM, 
	SIP_PARM_CALLING_PARTY_NUM, SIP_PARM_CHARGE_NUMBER, SIP_PARM_ORIG_LINE_INFO, 
	SIP_PARM_GENERIC_ADDR, SIP_PARM_GENERIC_DIGITS, SIP_PARM_GENERIC_NAME, SIP_PARM_JIP, 
	SIP_PARM_LOCAL_SERVICE_PROVIDER_IDENTIFICATION, -1};
	
static int sip_acm_params[] = {SIP_PARM_BACKWARD_CALL_IND, -1};
static int sip_faa_params[] = {SIP_PARM_FACILITY_IND, SIP_PARM_CALL_REF, -1};
static int sip_far_params[] = {SIP_PARM_FACILITY_IND, SIP_PARM_CALL_REF, -1};
static int sip_anm_params[] = { -1};
static int sip_con_params[] = {SIP_PARM_BACKWARD_CALL_IND, -1};
static int sip_rel_params[] = {SIP_PARM_CAUSE, -1};
static int sip_greset_params[] = {SIP_PARM_RANGE_AND_STATUS, -1};
static int sip_cot_params[] = {SIP_PARM_CONTINUITY_IND, -1};
static int sip_cpg_params[] = {SIP_PARM_EVENT_INFO, -1};
static int sip_cicgroup_params[] = {SIP_PARM_CIRCUIT_GROUP_SUPERVISION_IND, SIP_PARM_RANGE_AND_STATUS, -1};
static int sip_cqr_params[] = {SIP_PARM_RANGE_AND_STATUS, SIP_PARM_CIRCUIT_STATE_IND, -1};
static int sip_empty_params[] = { -1};

static struct sip_isup_message_data {
        int messagetype;
        int mand_fixed_params;
        int mand_var_params;
        int opt_params;
        int *param_list;
} sip_isup_messages[] = {
        {SIP_IAM, 4, 1, 1, sip_iam_params},
        {SIP_ACM, 1, 0, 1, sip_acm_params},
        {SIP_ANM, 0, 0, 1, sip_anm_params},
        {SIP_CON, 1, 0, 1, sip_con_params},
        {SIP_REL, 0, 1, 1, sip_rel_params},
        {SIP_RLC, 0, 0, 1, sip_empty_params},
        {SIP_GRS, 0, 1, 0, sip_greset_params},
        {SIP_GRA, 0, 1, 0, sip_greset_params},
        {SIP_CGB, 1, 1, 0, sip_cicgroup_params},
        {SIP_CGU, 1, 1, 0, sip_cicgroup_params},
        {SIP_CGBA, 1, 1, 0, sip_cicgroup_params},
        {SIP_CGUA, 1, 1, 0, sip_cicgroup_params},
        {SIP_COT, 1, 0, 0, sip_cot_params},
        {SIP_CCR, 0, 0, 0, sip_empty_params},
        {SIP_BLO, 0, 0, 0, sip_empty_params},
        {SIP_LPA, 0, 0, 0, sip_empty_params},
        {SIP_UBL, 0, 0, 0, sip_empty_params},
        {SIP_BLA, 0, 0, 0, sip_empty_params},
        {SIP_UBA, 0, 0, 0, sip_empty_params},
        {SIP_RSC, 0, 0, 0, sip_empty_params},
        {SIP_CVR, 0, 0, 0, sip_empty_params},
        {SIP_CVT, 0, 0, 0, sip_empty_params},
        {SIP_CPG, 1, 0, 1, sip_cpg_params},
        {SIP_UCIC, 0, 0, 0, sip_empty_params},
        {SIP_CQM, 0, 1, 0, sip_greset_params},
        {SIP_CQR, 0, 2, 0, sip_cqr_params},
        {SIP_FAA, 1, 0, 1, sip_faa_params},
        {SIP_FAR, 1, 0, 1, sip_far_params},
        {SIP_CFN, 0, 1, 0, sip_rel_params}
};

/* ISUP Parameter Pseudo-type */
struct sip_isup_parm_opt {
	unsigned char type;
	unsigned char len;
	unsigned char data[0];
};

struct isup_call {
	char called_party_num[ISUP_MAX_NUM];
	unsigned char called_nai;
	char calling_party_num[ISUP_MAX_NUM];
	unsigned char calling_party_cat;
	unsigned char calling_nai;
	unsigned char presentation_ind;
	unsigned char screening_ind;
	char charge_number[ISUP_MAX_NUM];
	unsigned char charge_nai;
	unsigned char charge_num_plan;
	unsigned char gen_add_num_plan;
	unsigned char gen_add_nai;
	char gen_add_number[ISUP_MAX_NUM];
	unsigned char gen_add_pres_ind;
	unsigned char gen_add_type;
	char gen_dig_number[ISUP_MAX_NUM];
	unsigned char gen_dig_type;
	unsigned char gen_dig_scheme;
	char jip_number[ISUP_MAX_NUM];
	unsigned char lspi_type;
	unsigned char lspi_scheme;
	unsigned char lspi_context;
	unsigned char lspi_spare;
	char lspi_ident[ISUP_MAX_NUM];
	int oli_ani2;
	unsigned int call_ref_ident;
	unsigned int call_ref_pc;
	char orig_called_num[ISUP_MAX_NUM];
	unsigned char orig_called_nai;
	unsigned char orig_called_pres_ind;
	unsigned char orig_called_screening_ind;
	char redirecting_num[ISUP_MAX_NUM];
	unsigned char redirecting_num_nai;
	unsigned char redirecting_num_presentation_ind;
	unsigned char redirecting_num_screening_ind;
	unsigned char generic_name_typeofname;
	unsigned char generic_name_avail;
	unsigned char generic_name_presentation;
	char generic_name[ISUP_MAX_NAME];
	int range;
	unsigned char status[255];
	int transcap;
	int l1prot;
	int cause;
	int causecode;
	int causeloc;
	int cot_check_passed;
	int cot_check_required;
	int cicgroupsupervisiontype;
	unsigned char event_info;
	unsigned short cic;
	unsigned char sls;
	struct isup_call *next;
	/* set DPC according to CIC's DPC, not linkset */
	unsigned int dpc;
	/* Backward Call Indicator variables */
	unsigned char called_party_status_ind;
};

